@@ -1,6 +1,7 @@
 # https://confluence.iter.org/display/IMP/UDA+data+mapping+tutorial
 
-import urllib
+import subprocess
+import shutil
 from .omas_utils import *
 from .omas_core import ODS, dynamic_ODS, omas_environment, omas_info_node, imas_json_dir, omas_rcparams
 from .omas_physics import cocos_signals
@@ -16,65 +17,43 @@ __all__ = [
     'mdsvalue',
 ]
 
+url_dir = os.sep.join([omas_rcparams['tmp_omas_dir'], 'machine_mappings', '{branch}', 'omas_machine_mappings_url_{branch}'])
+
 _python_tdi_namespace = {}
 
-url_dir = os.sep.join([omas_rcparams['tmp_omas_dir'], 'machine_mappings', '{branch}', 'omas_machine_mappings_url'])
-
-machine_expression_types = ['VALUE', 'ENVIRON', 'PYTHON', 'TDI', 'eval2TDI']
-
 
 def python_tdi_namespace(branch):
     '''
     Returns the namespace of the python_tdi.py file
-    This is done in such complicated way to allow `inspect.getsource` to work
 
     :param branch: remote branch to load
 
     :return: namespace
     '''
     # return cached python tdi function namespace
-    if _python_tdi_namespace.get('__branch__', False) == branch:
-        return _python_tdi_namespace
-
-    # clear python tdi function namespace
-    _python_tdi_namespace.clear()
-
-    # remove old machine_mappings modules
-    if 'omas.machine_mappings' in sys.modules:
-        del sys.modules['omas.machine_mappings']
-    if 'omas.machine_mappings.python_tdi' in sys.modules:
-        del sys.modules['omas.machine_mappings.python_tdi']
+    if branch in _python_tdi_namespace:
+        return _python_tdi_namespace[branch]
+    _python_tdi_namespace[branch] = {}
 
     # get local mapping functions
-    if branch is None:
-        exec('from omas.machine_mappings.python_tdi import *', _python_tdi_namespace)
+    if not branch:
+        exec('from omas.machine_mappings.python_tdi import *', _python_tdi_namespace[branch])
+
     # get mapping functions from GitHub
     else:
         printd(f'omas python tdi mappings from branch: `{branch}`', topic='machine')
 
-        # setup temporary import directory
-        dir = url_dir.format(branch=branch)
-        if not os.path.exists(dir):
-            os.makedirs(dir)
-        open(dir + os.sep + '__init__.py', 'w').close()
-
-        # download the python_tdi.py and write it to file
-        url = f"https://raw.githubusercontent.com/gafusion/omas/{branch}/omas/machine_mappings/python_tdi.py"
-        contents = urllib.request.urlopen(url).read().decode("utf-8")
-        with open(dir + os.sep + 'python_tdi.py', 'w') as f:
-            f.write(contents)
+        # make sure remote branch is transfered
+        machines(None, branch)
 
         # import from temporary directory
+        dir = url_dir.format(branch=branch)
         if dir + os.sep + '..' not in sys.path:
             sys.path.insert(0, dir + os.sep + '..')
-        exec('from omas_machine_mappings_url.python_tdi import *', _python_tdi_namespace)
 
-        # alias
-        sys.modules['omas.machine_mappings'] = sys.modules['omas_machine_mappings_url']
-        sys.modules['omas.machine_mappings.python_tdi'] = sys.modules['omas_machine_mappings_url.python_tdi']
+        exec(f'from omas_machine_mappings_url_{branch}.python_tdi import *', _python_tdi_namespace[branch])
 
-    _python_tdi_namespace['__branch__'] = branch
-    return _python_tdi_namespace
+    return _python_tdi_namespace[branch]
 
 
 def update_mapping(machine, location, value, cocosio=None, default_options=None, update_path=False):
@@ -115,7 +94,7 @@ def update_mapping(machine, location, value, cocosio=None, default_options=None,
 
     # add definition for new/updated location and update the .json file
     new_raw_mappings[ulocation] = value
-    filename, branch = machines(machine, None)
+    filename = machines(machine, '')
     with open(filename, 'w') as f:
         json.dump(new_raw_mappings, f, indent=1, separators=(',', ': '), sort_keys=True)
     print(f"Updated {machine} mapping for {ulocation}")
@@ -123,6 +102,8 @@ def update_mapping(machine, location, value, cocosio=None, default_options=None,
     # add the same call for arrays of structures going upstream
     if update_path:
         for uloc in [':'.join(ulocation.split(':')[: k + 1]) + ':' for k, l in enumerate(ulocation.split(':')[:-1])]:
+            if uloc in new_raw_mappings:
+                continue
             if 'COCOSIO' in value:
                 value = copy.copy(value)
                 del value['COCOSIO']
@@ -172,11 +153,16 @@ class mdsvalue(dict):
     '''
 
     def __init__(self, server, treename, pulse, TDI):
-        self.machine = server
         self.treename = treename
         self.pulse = pulse
         self.TDI = TDI
-        self.server = mds_machine_to_server_mapping(machine_mappings(self.machine, None)['__mdsserver__'], self.treename)
+        try:
+            # handle the case that server is just the machine name
+            server = machine_mappings(server, None)['__mdsserver__']
+        except NotImplementedError:
+            if '.' not in server:
+                raise
+        self.server = mds_machine_to_server_mapping(server, self.treename)
 
     def data(self):
         return self.raw(f'data({self.TDI})')
@@ -213,32 +199,35 @@ class mdsvalue(dict):
         if TDI is None:
             TDI = self.TDI
 
-        for fallback in [0, 1]:
-            if (self.server, self.treename, self.pulse) not in _mds_connection_cache:
-                conn = MDSplus.Connection(self.server)
-                if self.treename is not None:
-                    conn.openTree(self.treename, self.pulse)
-                _mds_connection_cache[(self.server, self.treename, self.pulse)] = conn
-            try:
-                conn = _mds_connection_cache[(self.server, self.treename, self.pulse)]
-                break
-            except Exception:
-                if (self.server, self.treename, self.pulse) in _mds_connection_cache:
-                    del _mds_connection_cache[(self.server, self.treename, self.pulse)]
-                if fallback:
-                    raise
-
         try:
+            for fallback in [0, 1]:
+                if (self.server, self.treename, self.pulse) not in _mds_connection_cache:
+                    conn = MDSplus.Connection(self.server)
+                    if self.treename is not None:
+                        conn.openTree(self.treename, self.pulse)
+                    _mds_connection_cache[(self.server, self.treename, self.pulse)] = conn
+                try:
+                    conn = _mds_connection_cache[(self.server, self.treename, self.pulse)]
+                    break
+                except Exception as _excp:
+                    if (self.server, self.treename, self.pulse) in _mds_connection_cache:
+                        del _mds_connection_cache[(self.server, self.treename, self.pulse)]
+                    if fallback:
+                        raise
+
             return MDSplus.Data.data(conn.get(TDI))
         except Exception as _excp:
             txt = []
-            for item in ['machine', 'server', 'treename', 'pulse']:
+            for item in ['server', 'treename', 'pulse']:
                 txt += [f' - {item}: {getattr(self, item)}']
             txt += [f' - TDI: {TDI}']
             raise _excp.__class__(str(_excp) + '\n' + '\n'.join(txt))
 
 
-def machine_to_omas(ods, machine, pulse, location, options={}, branch=None, user_machine_mappings=None, cache=None):
+machine_expression_types = ['VALUE', 'ENVIRON', 'PYTHON', 'TDI', 'eval2TDI']
+
+
+def machine_to_omas(ods, machine, pulse, location, options={}, branch='', user_machine_mappings=None, cache=None):
     '''
     Routine to convert machine data to ODS
 
@@ -264,11 +253,18 @@ def machine_to_omas(ods, machine, pulse, location, options={}, branch=None, user
         user_machine_mappings = {}
 
     location = l2o(p2l(location))
-    mappings = machine_mappings(machine, branch, user_machine_mappings)
-    options_with_defaults = copy.copy(mappings['__options__'])
-    options_with_defaults.update(options)
-    options_with_defaults.update({'machine': machine, 'pulse': pulse, 'location': location})
-    mapped = mappings[location]
+
+    for branch in [branch, 'master']:
+        mappings = machine_mappings(machine, branch, user_machine_mappings)
+        options_with_defaults = copy.copy(mappings['__options__'])
+        options_with_defaults.update(options)
+        options_with_defaults.update({'machine': machine, 'pulse': pulse, 'location': location})
+        try:
+            mapped = mappings[location]
+        except KeyError:
+            if branch == 'master':
+                raise
+    idm = (machine, branch)
 
     # cocosio
     cocosio = None
@@ -295,19 +291,21 @@ def machine_to_omas(ods, machine, pulse, location, options={}, branch=None, user
             ods = cache[call]
         else:
             namespace = {}
-            namespace.update(_namespace_mappings[machine])
+            namespace.update(_namespace_mappings[idm])
             namespace['ods'] = ODS()
-            namespace['__file__'] = machines(machine, branch)[0][:-5] + '.py'
-            print(machines(machine, branch)[0][:-5] + '.py')
-            tmp = compile(call, machines(machine, branch)[0][:-5] + '.py', 'exec')
+            namespace['__file__'] = machines(machine, branch)[:-5] + '.py'
+            tmp = compile(call, machines(machine, branch)[:-5] + '.py', 'exec')
             exec(tmp, namespace)
             ods = namespace[mapped.get('RETURN', 'ods')]
             if isinstance(cache, dict):
                 cache[call] = ods
         if location.endswith(':'):
-            return int(len(ods[u2n(location[:-2], [0] * 100)])), {'raw_data': ods, 'processed_data': ods, 'cocosio': cocosio}
+            return (
+                int(len(ods[u2n(location[:-2], [0] * 100)])),
+                {'raw_data': ods, 'processed_data': ods, 'cocosio': cocosio, 'branch': mappings['__branch__']},
+            )
         else:
-            return ods, {'raw_data': ods, 'processed_data': ods, 'cocosio': cocosio}
+            return ods, {'raw_data': ods, 'processed_data': ods, 'cocosio': cocosio, 'branch': mappings['__branch__']}
 
     # MDS+
     elif 'TDI' in mapped:
@@ -326,7 +324,7 @@ def machine_to_omas(ods, machine, pulse, location, options={}, branch=None, user
 
     # handle size definition for array of structures
     if location.endswith(':'):
-        return int(data), {'raw_data': data0, 'processed_data': data, 'cocosio': cocosio}
+        return int(data), {'raw_data': data0, 'processed_data': data, 'cocosio': cocosio, 'branch': mappings['__branch__']}
 
     # transpose manipulation
     if mapped.get('TRANSPOSE', False):
@@ -361,7 +359,7 @@ def machine_to_omas(ods, machine, pulse, location, options={}, branch=None, user
                 for k in range(data.shape[0]):
                     ods[u2n(location, [k] + [0] * 10)] = nanfilter(data[k, ...])
 
-    return ods, {'raw_data': data0, 'processed_data': data, 'cocosio': cocosio}
+    return ods, {'raw_data': data0, 'processed_data': data, 'cocosio': cocosio, 'branch': mappings['__branch__']}
 
 
 def machine_mapping_function(__all__):
@@ -446,7 +444,7 @@ def load_omas_machine(
     consistency_check=True,
     imas_version=omas_rcparams['default_imas_version'],
     cls=ODS,
-    branch=None,
+    branch='',
     user_machine_mappings=None,
 ):
     printd('Loading from %s' % machine, topic='machine')
@@ -465,7 +463,7 @@ class dynamic_omas_machine(dynamic_ODS):
     This class is not to be used by itself, but via the ODS.open() method.
     """
 
-    def __init__(self, machine, pulse, options={}, branch=None, user_machine_mappings=None, verbose=True):
+    def __init__(self, machine, pulse, options={}, branch='', user_machine_mappings=None, verbose=True):
         self.kw = {'machine': machine, 'pulse': pulse, 'options': options, 'branch': branch, 'user_machine_mappings': user_machine_mappings}
         self.active = False
         self.cache = {}
@@ -527,7 +525,7 @@ class dynamic_omas_machine(dynamic_ODS):
 _machines_dict = {}
 
 
-def machines(machine=None, branch=None):
+def machines(machine=None, branch=''):
     '''
     Function to get machines that have their mappings defined
     This function takes care of remote transfer the needed files (both .json and .py) if a remote branch is requested
@@ -539,52 +537,49 @@ def machines(machine=None, branch=None):
     :return: if `machine==None` returns dictionary with list of machines and their json mapping files
              if `machine` is a string, then returns json mapping filename and branch
     '''
-    # return cached results
-    if '__branch__' in _machines_dict and _machines_dict['__branch__'] == branch:
-        if machine is None:
-            return _machines_dict
-        elif machine in _machines_dict:
-            return _machines_dict[machine]
 
-    # local machines
-    for filename in glob.glob(imas_json_dir + '/../machine_mappings/*.json'):
-        _machines_dict[os.path.splitext(os.path.split(filename)[1])[0]] = os.path.abspath(filename)
+    # return cached results
+    if branch in _machines_dict:
+        if machine is None:
+            return _machines_dict[branch]
+        elif machine in _machines_dict:
+            return _machines_dict[branch][machine]
+
+    # local mappings
+    if not branch:
+        dir = imas_json_dir + '/../machine_mappings'
+
+    # remote mappings from GitHub
+    else:
+        if branch == 'master':
+            svn_branch = 'trunk'
+        else:
+            svn_branch = 'branches/' + branch
+
+        dir = url_dir.format(branch=branch)
+        if os.path.exists(dir):
+            shutil.rmtree(dir)
+        subprocess.Popen(
+            f'''
+svn export --force https://github.com/gafusion/omas.git/{svn_branch}/omas/machine_mappings/ {dir}
+''',
+            stdout=subprocess.PIPE,
+            shell=True,
+        ).communicate()[0]
+
+    # go through machine files
+    _machines_dict[branch] = {}
+    for filename in glob.glob(f'{dir}/*.json'):
+        _machines_dict[branch][os.path.splitext(os.path.split(filename)[1])[0]] = os.path.abspath(filename)
 
     # return list of supported machines
     if machine is None:
-        if branch is None:
-            return _machines_dict
-        else:
-            raise NotImplementedError(f'Cannot list machine mappings on GitHub branches')
-
+        return _machines_dict[branch]
     # return filename with mappings for this machine
     else:
-        # try `master` branch if not in local
-        if machine not in _machines_dict and branch is None:
-            branch = 'master'
-
-        # get remote mappings
-        if branch is not None:
-
-            # setup temporary machine_mappings directory
-            dir = url_dir.format(branch=branch)
-            if not os.path.exists(dir):
-                os.makedirs(dir)
-
-            # download machine.json/py files and write them to file
-            for ext in ['json', 'py']:
-                try:
-                    url = f"https://raw.githubusercontent.com/gafusion/omas/{branch}/omas/machine_mappings/{machine}.json"
-                    contents = urllib.request.urlopen(url).read().decode("utf-8")
-                    filename = dir + os.sep + f'{machine}.{ext}'
-                    with open(filename, 'w') as f:
-                        f.write(contents)
-                    _machines_dict[machine] = filename
-                except Exception as _excp:
-                    if ext != 'json':
-                        printd(f'No machine mappings for `{machine}` from branch `{branch}`: {repr(_excp)}', topic='machine')
-                        raise NotImplementedError(f'No machine mapping for `{machine}`. Valid machines are: {list(_machines_dict.keys())}')
-        return _machines_dict[machine], branch
+        if machine not in _machines_dict[branch]:
+            raise NotImplementedError(f'Machine `{machine}` has no mapping defined')
+        return _machines_dict[branch][machine]
 
 
 _machine_mappings = {}
@@ -607,16 +602,16 @@ def machine_mappings(machine, branch, user_machine_mappings=None, return_raw_map
     if user_machine_mappings is None:
         user_machine_mappings = {}
 
+    idm = (machine, branch)
+
     if (
         return_raw_mappings
         or machine not in _machine_mappings
-        or list(_user_machine_mappings.keys()) + list(user_machine_mappings.keys())
-        != _machine_mappings[machine]['__user_machine_mappings__']
+        or list(_user_machine_mappings.keys()) + list(user_machine_mappings.keys()) != _machine_mappings[idm]['__user_machine_mappings__']
     ):
 
         # figure out mapping file
-        # this function will take care of remote transfer the needed files (both .json and .py) if a remote branch is requested
-        filename, branch = machines(machine, branch)
+        filename = machines(machine, branch)
 
         # load mappings from file
         with open(filename, 'r') as f:
@@ -638,11 +633,14 @@ def machine_mappings(machine, branch, user_machine_mappings=None, return_raw_map
             mappings.pop('__user_machine_mappings__')
             return mappings
 
+        mappings['__filename__'] = filename
+        mappings['__branch__'] = branch
+
         # read the machine specific python mapping functions
-        _namespace_mappings[machine] = {}
+        _namespace_mappings[idm] = {}
         if os.path.exists(os.path.splitext(filename)[0] + '.py'):
             with open(os.path.splitext(filename)[0] + '.py', 'r') as f:
-                exec(f.read(), _namespace_mappings[machine])
+                exec(f.read(), _namespace_mappings[idm])
 
         # generate TDI for cocos_rules
         for item in mappings['__cocos_rules__']:
@@ -689,6 +687,6 @@ def machine_mappings(machine, branch, user_machine_mappings=None, return_raw_map
                 raise ValueError(f'{location} should not have COCOS specified, or COCOS definition should be added to omas_cocos file')
 
         # cache
-        _machine_mappings[machine] = mappings
+        _machine_mappings[idm] = mappings
 
-    return _machine_mappings[machine]
+    return _machine_mappings[idm]
