@@ -79,12 +79,12 @@ def consistent_times(self, attempt_fix=True, raise_errors=True):
 
     # if called at top level, loop over all data structures
     if not len(self.location):
-        out = []
+        out = {}
         for ds in self:
-            out.append(self.getraw(ds).physics_consistent_times(attempt_fix=attempt_fix, raise_errors=raise_errors))
-        if any([k is False for k in out]):
+            out[ds] = self.getraw(ds).physics_consistent_times(attempt_fix=attempt_fix, raise_errors=raise_errors)
+        if any([k is False for k in out.values()]):
             return False
-        elif any([k is None for k in out]):
+        elif any([k is None for k in out.values()]):
             return None
         else:
             return True
@@ -219,6 +219,7 @@ def summary_greenwald(ods, update=True):
 
     :return: updated ods
     """
+
     ods_n = ods
     if not update:
         from omas import ODS
@@ -240,6 +241,100 @@ def summary_greenwald(ods, update=True):
     return ods_n
 
 
+@add_to__ODS__
+@preprocess_ods('core_profiles', 'equilibrium')
+def summary_lineaverage_density(ods, line_grid=2000, time_index=None, update=True):
+
+    """
+    Calculates line-average electron density for each time slice and stores them in the summary ods
+
+    :param ods: input ods
+    
+    :param line_grid: number of points to calculate line average density over (includes point outside of boundary)
+    
+    :param time_index: time slices to process
+
+    :param update: operate in place
+
+    :return: updated ods
+    """
+    import scipy
+
+    ods_n = ods
+    if not update:
+        from omas import ODS
+
+        ods_n = ODS().copy_attrs_from(ods)
+
+    if time_index is None:
+        for time_index in range(len(ods['core_profiles']['profiles_1d'])):
+            summary_lineaverage_density(ods_n, line_grid=line_grid, time_index=time_index, update=False)
+        return ods_n
+
+    Rb = ods['equilibrium']['time_slice'][time_index]['boundary']['outline']['r']
+    Zb = ods['equilibrium']['time_slice'][time_index]['boundary']['outline']['z']
+
+    Rgrid = ods['equilibrium']['time_slice'][time_index]['profiles_2d'][0]['grid']['dim1']
+    Zgrid = ods['equilibrium']['time_slice'][time_index]['profiles_2d'][0]['grid']['dim2']
+
+    psi2d = ods['equilibrium']['time_slice'][time_index]['profiles_2d'][0]['psi']
+    psi_interp = scipy.interpolate.interp2d(Zgrid, Rgrid, psi2d)
+    psi_eq = ods['equilibrium']['time_slice'][time_index]['profiles_1d']['psi']
+    rhon_eq = ods['equilibrium']['time_slice'][time_index]['profiles_1d']['rho_tor_norm']
+    rhon_cp = ods['core_profiles']['profiles_1d'][time_index]['grid']['rho_tor_norm']
+    ne = ods['core_profiles']['profiles_1d'][time_index]['electrons']['density_thermal']
+    ne = numpy.interp(rhon_eq, rhon_cp, ne)
+    tck = scipy.interpolate.splrep(psi_eq, ne, k=3)
+
+    if 'time' not in ods['interferometer']:
+        ods_n['interferometer.ids_properties.homogeneous_time'] = 1
+        ods_n['interferometer']['time'] = copy.copy(ods['core_profiles']['time'])
+
+    ifpaths = [['first_point', 'second_point'], ['second_point', 'third_point']]
+    for channel in ods['interferometer']['channel']:
+        ne_line_paths = []
+        dist_paths = []
+        for ifpath in ifpaths:
+            R1 = ods['interferometer']['channel'][channel]['line_of_sight'][ifpath[0]]['r']
+            Z1 = ods['interferometer']['channel'][channel]['line_of_sight'][ifpath[0]]['z']
+            phi1 = ods['interferometer']['channel'][channel]['line_of_sight'][ifpath[0]]['phi']
+            x1 = R1 * numpy.cos(phi1)
+            y1 = R1 * numpy.sin(phi1)
+
+            R2 = ods['interferometer']['channel'][channel]['line_of_sight'][ifpath[1]]['r']
+            Z2 = ods['interferometer']['channel'][channel]['line_of_sight'][ifpath[1]]['z']
+            phi2 = ods['interferometer']['channel'][channel]['line_of_sight'][ifpath[1]]['phi']
+            x2 = R2 * numpy.cos(phi2)
+            y2 = R2 * numpy.sin(phi2)
+
+            xline = numpy.linspace(x1, x2, line_grid)
+            yline = numpy.linspace(y1, y2, line_grid)
+            Rline = numpy.linspace(R1, R2, line_grid)
+            Zline = numpy.linspace(Z1, Z2, line_grid)
+            dist = numpy.zeros(line_grid)
+
+            for i, Rval in enumerate(Rline):
+                dist[i] = numpy.min((Rline[i] - Rb) ** 2 + (Zline[i] - Zb) ** 2)
+
+            zero_crossings = numpy.where(numpy.diff(numpy.sign(numpy.gradient(dist))))[0]
+            i1 = zero_crossings[0]
+            i2 = zero_crossings[-1]
+
+            psival = [psi_interp(Zline[i], Rline[i])[0] for i in range(i1, i2, numpy.sign(i2 - i1))]
+            ne_interp = scipy.interpolate.splev(psival, tck)
+            ne_line = numpy.trapz(ne_interp)
+            ne_line /= abs(i2 - i1)
+            ne_line_paths.append(ne_line)
+            dist_paths.append(numpy.sqrt((xline[i2] - xline[i1]) ** 2 + (yline[i2] - yline[i1]) ** 2 + (Zline[i2] - Zline[i1]) ** 2))
+
+        ne_line = numpy.average(ne_line_paths, weights=dist_paths)
+        if F'interferometer.channel.{channel}.n_e_line_average.data' not in ods_n:
+            ods_n['interferometer']['channel'][channel]['n_e_line_average']['data'] = numpy.zeros(len(ods_n['interferometer']['time']))
+
+        ods_n['interferometer']['channel'][channel]['n_e_line_average']['data'][time_index] = ne_line
+    return ods_n
+
+
 @add_to__ODS__
 @preprocess_ods('core_profiles', 'core_sources', 'equilibrium')
 def summary_taue(ods, update=True):
@@ -383,7 +478,7 @@ def summary_total_powers(ods, update=True):
 
     for key, value in power_dict.items():
         if numpy.sum(value) > 0:
-            if key is 'total':
+            if key == 'total':
                 ods_n['summary.global_quantities.power_steady.value'] = numpy.array(value)
                 continue
             ods_n[f'summary.heating_current_drive.{key}[0].power.value'] = numpy.array(value)
@@ -470,8 +565,8 @@ def summary_consistent_global_quantities(ods, ds=None, update=True):
 def core_profiles_consistent(ods, update=True, use_electrons_density=False):
     """
     Calls all core_profiles consistency functions including
-      - core_profiles_pressures
       - core_profiles_densities
+      - core_profiles_pressures
       - core_profiles_zeff
 
     :param ods: input ods
@@ -484,8 +579,8 @@ def core_profiles_consistent(ods, update=True, use_electrons_density=False):
 
     :return: updated ods
     """
-    ods = core_profiles_pressures(ods, update=update)
-    core_profiles_densities(ods)
+    ods = core_profiles_densities(ods, update=update)
+    core_profiles_pressures(ods)
     core_profiles_zeff(ods, use_electrons_density=use_electrons_density)
     return ods
 
@@ -1206,6 +1301,7 @@ def grids_ggd_points_triangles(grid):
 
     :return: tuple with points and triangles
     """
+    # objects_per_dimension: 0 = nodes, 1 = edges, 2 = faces, 3 = cells / volumes
     points = grid['space[0].objects_per_dimension[0].object[:].geometry']
     triangles = grid['space[0].objects_per_dimension[2].object[:].nodes']
     return points, triangles
@@ -1279,6 +1375,24 @@ def scatter_to_rectangular(r, z, data, R, Z, method=['nearest', 'linear', 'cubic
     return R[0, :], Z[:, 0], intepolated_data
 
 
+@add_to__ODS__
+def check_iter_scenario_requirements(ods):
+    '''
+    Check that the current ODS satisfies the ITER scenario database requirements as defined in https://confluence.iter.org/x/kQqOE
+
+    :return: list of elements that are missing to satisfy the ITER scenario requirements
+    '''
+    from .omas_imas import iter_scenario_requirements
+
+    fail = []
+    for item in iter_scenario_requirements:
+        try:
+            ods[item]  # acccessing a leaf that has no data will raise an error
+        except Exception:
+            fail.append(item)
+    return fail
+
+
 @add_to__ALL__
 def transform_current(rho, JtoR=None, JparB=None, equilibrium=None, includes_bootstrap=False):
     """
@@ -1665,7 +1779,9 @@ def identify_cocos(B0, Ip, q, psi, clockwise_phi=None, a=None):
 
 @add_to__ALL__
 @contextmanager
-def omas_environment(ods, cocosio=None, coordsio=None, unitsio=None, input_data_process_functions=None, xmlcodeparams=False, **kw):
+def omas_environment(
+    ods, cocosio=None, coordsio=None, unitsio=None, input_data_process_functions=None, xmlcodeparams=False, dynamic_path_creation=None, **kw
+):
     """
     Provides environment for data input/output to/from OMAS
 
@@ -1679,6 +1795,11 @@ def omas_environment(ods, cocosio=None, coordsio=None, unitsio=None, input_data_
 
     :param xmlcodeparams: view code.parameters as an XML string while in this environment
 
+    :param dynamic_path_creation: whether to dynamically create the path when setting an item
+                                  * False: raise an error when trying to access a structure element that does not exists
+                                  * True (default): arrays of structures can be incrementally extended by accessing at the next element in the array
+                                  * 'dynamic_array_structures': arrays of structures can be dynamically extended
+
     :param kw: extra keywords set attributes of the ods (eg. 'consistency_check', 'dynamic_path_creation', 'imas_version')
 
     :return: ODS with environment set
@@ -1689,8 +1810,7 @@ def omas_environment(ods, cocosio=None, coordsio=None, unitsio=None, input_data_
         from omas import ODS
 
         tmp = ODS(cocos=ods.cocos)
-        tmp.dynamic_path_creation = 'dynamic_array_structures'
-        with omas_environment(tmp, cocosio=cocosio):
+        with omas_environment(tmp, cocosio=cocosio, dynamic_path_creation='dynamic_array_structures'):
             tmp.update(coordsio)
         coordsio = tmp
 
@@ -1698,6 +1818,7 @@ def omas_environment(ods, cocosio=None, coordsio=None, unitsio=None, input_data_
         raise ValueError('cocosio can only be an integer')
 
     # backup attributes
+    bkp_dynamic_path_creation = omas_rcparams['dynamic_path_creation']
     bkp_cocosio = ods.cocosio
     bkp_coordsio = ods.coordsio
     bkp_unitsio = ods.unitsio
@@ -1714,6 +1835,8 @@ def omas_environment(ods, cocosio=None, coordsio=None, unitsio=None, input_data_
         ods.coordsio = (ods, coordsio)
     if unitsio is not None:
         ods.unitsio = unitsio
+    if dynamic_path_creation is not None:
+        omas_rcparams['dynamic_path_creation'] = dynamic_path_creation
 
     # set input_data_process_functions
     if input_data_process_functions is not None:
@@ -1738,6 +1861,7 @@ def omas_environment(ods, cocosio=None, coordsio=None, unitsio=None, input_data_
         if xmlcodeparams:
             ods.codeparams2dict()
         # restore attributes
+        omas_rcparams['dynamic_path_creation'] = bkp_dynamic_path_creation
         ods.cocosio = bkp_cocosio
         ods.coordsio = bkp_coordsio
         ods.unitsio = bkp_unitsio
@@ -1806,7 +1930,7 @@ def generate_cocos_signals(structures=[], threshold=0, write=True, verbose=False
     # if generate_cocos_signals is run after omas has been used for something else
     # (eg. when running test_examples) it may be that _extra_structures is not empty
     # Thus, we clear _structures and _extra_structures to make sure that
-    # dict_structures() is not polluted by _extra_structures
+    # structures_filenames() is not polluted by _extra_structures
     _structures_bkp = copy.deepcopy(_structures)
     _extra_structures_bkp = copy.deepcopy(_extra_structures)
     try:
@@ -1851,7 +1975,7 @@ def generate_cocos_signals(structures=[], threshold=0, write=True, verbose=False
 #        transforms['BP'] = transforms['POL']
 #        transforms[None] = 1
 #
-                
+
 _cocos_signals = {}
 """
         ]
@@ -1887,10 +2011,9 @@ _cocos_signals = {}
 
             out[structure] = {}
             ods[structure]
-            d = dict_structures(imas_version=omas_rcparams['default_imas_version'])
             m = 0
             # generate score and add reason for scoring
-            for item in sorted(list(_structures[(structure, omas_rcparams['default_imas_version'])].keys())):
+            for item in sorted(list(load_structure(structure, omas_rcparams['default_imas_version'])[0].keys())):
                 item = i2o(item)
                 item_ = item
                 if any([item.endswith(k) for k in [':.values', ':.value', ':.data']]):
