@@ -167,6 +167,25 @@ def core_profiles(ods, time_index=0, add_junk_ion=False, include_pressure=True):
     return ods
 
 
+@add_to_ODS
+def ic_antennas(ods):
+    """
+    Add sample ic_antennas data
+    This method operates in in-place.
+
+    :param ods: ODS instance
+
+    :param time_index: int
+
+    :return: ODS instance with profiles added
+    """
+    from omas import load_omas_json
+
+    ods.update(load_omas_json(imas_json_dir + '/../samples/sample_ic_antennas_ods.json', consistency_check=False))
+
+    return ods
+
+
 @add_to_ODS
 def core_sources(ods, time_index=0):
     """
@@ -386,7 +405,7 @@ def charge_exchange(ods, nc=10):
 
     :param ods: ODS instance
 
-    :param nc: Number of channels to add.
+    :param nc: Number of channels to add
 
     :return: ODS instance with fake CER hardware information added
     """
@@ -402,7 +421,7 @@ def charge_exchange(ods, nc=10):
         ch['position.phi.data'] = numpy.array([6.5])
         ch['position.r.data'] = numpy.array([r[i]])
         ch['position.z.data'] = numpy.array([z[i]])
-
+    ods['charge_exchange.time'] = [0.0]
     return ods
 
 
@@ -579,42 +598,19 @@ def wall(ods):
 
     :return: ODS instance with added wall description
     """
+    ods['wall.time'] = [0.0]
     ods['wall.description_2d[0].limiter.type.description'] = 'first wall'
     ods['wall.description_2d[0].limiter.type.index'] = 0
     ods['wall.description_2d[0].limiter.type.name'] = 'first_wall'
     ods['wall.description_2d[0].limiter.unit[0].outline.r'] = [
-        1.0,
-        1.0,
-        1.3,
-        1.4,
-        1.6,
-        2.15,
-        2.35,
-        2.35,
-        2.15,
-        1.800,
-        1.350,
-        1.35,
-        1.10,
-        1.00,
-        1.0,
+        # fmt: off
+        1.0, 1.0, 1.3, 1.4, 1.6, 2.15, 2.35, 2.35, 2.15, 1.800, 1.350, 1.35, 1.10, 1.00, 1.0
+        # fmt: on
     ]
     ods['wall.description_2d[0].limiter.unit[0].outline.z'] = [
-        0.0,
-        1.4,
-        1.4,
-        1.3,
-        1.1,
-        1.00,
-        0.50,
-        -0.5,
-        -1.0,
-        -1.25,
-        -1.25,
-        -1.4,
-        -1.4,
-        -1.3,
-        0.0,
+        # fmt: off
+        0.0, 1.4, 1.4, 1.3, 1.1, 1.00, 0.50, -0.5, -1.0, -1.25, -1.25, -1.4, -1.4, -1.3, 0.0
+        # fmt: on
     ]
 
     return ods
