@@ -22,6 +22,11 @@ def set_omas_warnings():
     warnings.filterwarnings('always', category=DeprecationWarning, message='Using or importing the ABCs.*')
     warnings.filterwarnings('always', category=RuntimeWarning, message='invalid value encountered.*')
     warnings.filterwarnings('always', category=DeprecationWarning, message='please use dns.resolver.resolve.*')
+    warnings.filterwarnings(
+        'always', category=UserWarning, message='Matplotlib is currently using agg, which is a non-GUI backend, so cannot show the figure'
+    )
+    warnings.filterwarnings('always', category=DeprecationWarning, message='tostring\(\) is deprecated. Use tobytes\(\) instead.')
+
     print('OMAS warnings set to hard mode')
     return
 
